/****************************************************************
 Licensed Materials - Property of IBM and/or HCL 
(c) Copyright HCL Technologies Ltd. 2018.  All Rights Reserved.

 Note to U.S. Government Users Restricted Rights:
 Use, duplication or disclosure restricted by GSA ADP Schedule:
 Contract with IBM Corp.
*****************************************************************/

package com.hcl.test.rm.model.tools;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import com.hcl.test.rm.model.RMObservable;
import com.hcl.test.rm.model.RMObservableContainer;
import com.hcl.test.rm.model.RMObservableData;

public class RMModelLookupTools {

	/**
	 * Visitor interface to be used with {@link RMModelLookupTools#accept(RMObservable, IRMModelVisitor)}
	 */
	public static interface IRMModelVisitor {
		void visit(RMObservableContainer container);
		void visit(RMObservableData data);
	}

	/**
	 * Visitor interface to be used with {@link RMModelLookupTools#accept(RMObservable, IRMModelSearchVisitor)}
	 */
	public static interface IRMModelSearchVisitor {
		boolean visit(RMObservable observable);
	}

	/**
	 * Visitor interface to be used with {@link RMModelLookupTools#accept(RMObservableContainer, IRMModelCollectedParentsVisitor)}
	 */
	public static interface IRMModelCollectedParentsVisitor {
		void visit(RMObservableContainer container, List<RMObservableContainer> parents);
		void visit(RMObservableData data, List<RMObservableContainer> parents);
	}
	
	/**
	 * Visitor starting point that apply given {@link IRMModelVisitor} to each and every node in the observable tree.
	 * @param observable
	 * @param visitor
	 */
	public static void accept(RMObservable observable, IRMModelVisitor visitor) {
		if (observable instanceof RMObservableContainer) {
			RMObservableContainer container = (RMObservableContainer)observable;
			visitor.visit(container);
			if (container.getChildren() != null) {
				for (RMObservable children : container.getChildren()) {
					accept(children, visitor);
				}
			}
		} else {
			RMObservableData data = (RMObservableData)observable;
			visitor.visit(data);
		}
	}

	/**
	 * Search visitor starting point that apply given {@link IRMModelSearchVisitor} to each and every node in the observable tree as long as
	 * {@link IRMModelSearchVisitor#visit(RMObservable)} returns false and stops the visit when true is returned by the same.
	 * @param observable
	 * @param visitor
	 * @return true if the visitor returned true, false otherwise
	 */
	public static boolean accept(RMObservable observable, IRMModelSearchVisitor visitor) {
		if (visitor.visit(observable)) {
			return true;
		}
		if (observable instanceof RMObservableContainer) {
			RMObservableContainer container = (RMObservableContainer)observable;
			if (container.getChildren() != null) {
				for (RMObservable children : container.getChildren()) {
					if (accept(children, visitor)) {
						return true;
					}
				}
			}
		}
		return false;
	}

	/**
	 * Search visitor starting point that apply given {@link IRMModelCollectedParentsVisitor} to each and every node in the observable tree.
	 * @param observable
	 * @param visitor
	 * @return true if the visitor returned true, false otherwise
	 */
	public static void accept(RMObservableContainer container, IRMModelCollectedParentsVisitor visitor) {
		List<RMObservableContainer> parents = new LinkedList<>();
		accept(container, parents, visitor);
	}
	
	private static void accept(RMObservable observable, List<RMObservableContainer> parents, IRMModelCollectedParentsVisitor visitor) {
		if (observable instanceof RMObservableContainer) {
			RMObservableContainer container = (RMObservableContainer)observable;
			visitor.visit(container, parents);
			parents.add(container);
			if (container.getChildren() != null) {
				for (RMObservable children : container.getChildren()) {
					accept(children, parents, visitor);
				}
			}
			parents.remove(parents.size() - 1);
		} else {
			RMObservableData data = (RMObservableData)observable;
			visitor.visit(data, parents);
		}
	}
	
	/**
	 * Lookup across the observable tree looking for a given observable id
	 * @param root starting node
	 * @param observableId id to search for
	 * @return the matching observable or null if none
	 */
	public static RMObservable findObservable(RMObservable root, String observableId) {
		RMObservable[] result = new RMObservable[] { null };
		accept(root, new IRMModelSearchVisitor() {
			@Override
			public boolean visit(RMObservable observable) {
				if (observable.getId().equals(observableId)) {
					result[0] = observable;
					return true;
				}
				return false;
			}
		});
		return result[0];
	}

	/**
	 *Lookup across the observable tree looking for the given observable ids
	 * @param root starting node
	 * @param ids to search for
	 * @return the set of matching observable or empty if none
	 */
	public static Set<RMObservable> findObservables(RMObservable root, Collection<String> observableIds) {
		Set<RMObservable> result = new HashSet<>();
		accept(root, new IRMModelSearchVisitor() {
			@Override
			public boolean visit(RMObservable observable) {
				if (observableIds.contains(observable.getId())) {
					result.add(observable);
					return result.size() == observableIds.size(); // We're done if everything is found
				}
				return false;
			}
		});
		return result;
	}

	/**
	 * A quick lookup to the observable tree looking for the given observable data ids
	 * @param root starting node
	 * @param ids to search for
	 * @return the set of matching observable or empty if none
	 */
	public static Set<RMObservableData> findObservableDatas(RMObservable root, Collection<String> observableIds) {
		Set<RMObservableData> result = new LinkedHashSet<>();
		accept(root, new IRMModelSearchVisitor() {
			@Override
			public boolean visit(RMObservable observable) {
				if (observable instanceof RMObservableData && observableIds.contains(observable.getId())) {
					result.add((RMObservableData)observable);
					return result.size() == observableIds.size();
				}
				return false;
			}
		});
		return result;
	}

}
