/****************************************************************
 Licensed Materials - Property of IBM and/or HCL
(c) Copyright HCL Technologies Ltd. 2018.  All Rights Reserved.

 Note to U.S. Government Users Restricted Rights:
 Use, duplication or disclosure restricted by GSA ADP Schedule:
 Contract with IBM Corp.
*****************************************************************/
define(["require", "exports", "jquery", "jquery.form", "jquery.serializejson", "jquery.validate", "bootstrap"], function (require, exports, $) {
    "use strict";
    exports.__esModule = true;
    var FormHallManager = /** @class */ (function () {
        function FormHallManager(formSelector, handlers) {
            var _this = this;
            this.formSelector = formSelector;
            this.handlers = handlers;
            this.onShowListeners = [];
            this.updateDataListeners = [];
            this.dirty = false;
            this.keepMessages = false;
            this.$form = $(formSelector);
            this.$form.change(function () {
                _this.setDirty(true);
            }).keypress(function () {
                _this.setDirty(true);
            }).on('reset', function () {
                setTimeout(function () {
                    if (!_this.keepMessages) {
                        _this.hideMessageAreas();
                    }
                    else {
                        _this.keepMessages = false;
                    }
                    _this.validator['hideAllErrors'].call(_this.$form);
                    _this.onUpdateData();
                    _this.setDirty(false);
                });
            });
            this.installDefaultValues();
            this.adaptConstraints();
            this.adaptInputRanges();
            this.setupChoiceItemsUrls();
            this.installDepends();
        }
        FormHallManager.prototype.setHandlers = function (handlers) {
            this.handlers = handlers;
        };
        FormHallManager.prototype.installDefaultValues = function () {
            this.$form.find('[data-formhall-default-value][data-formhall-default-value!=""]').each(function (index, formElement) {
                var $formElement = $(formElement);
                var defaultValue = $formElement.data('formhallDefaultValue');
                if (defaultValue) {
                    if ($formElement.attr('type') === 'checkbox') {
                        $formElement.attr('checked', 'checked');
                    }
                    else { // almost all others like text, etc.
                        $formElement.attr('value', defaultValue);
                    }
                }
            });
            this.$form.find('[data-formhall-container-default-value][data-formhall-container-default-value!=""]').each(function (index, container) {
                var $container = $(container);
                var defaultValue = $container.data('formhallContainerDefaultValue');
                if (defaultValue) {
                    if ($container.prop('tagName') === 'SELECT') {
                        $container.find('option').each(function (index, option) {
                            var $option = $(option);
                            var value = $option.attr('value');
                            if (value === defaultValue) {
                                $option.attr('selected', 'selected');
                            }
                        });
                    }
                    else if ($container.data('formhallRole') === 'field-range') { // input range
                        var $output = $container.find('output');
                        $output.text(defaultValue);
                    }
                    else { // set of radio buttons
                        $container.find('input[type="radio"]').each(function (index, radio) {
                            var $radio = $(radio);
                            var value = $radio.attr('value');
                            if (value === defaultValue) {
                                $radio.attr('checked', 'checked');
                            }
                        });
                    }
                }
            });
        };
        FormHallManager.prototype.adaptConstraints = function () {
            var _this = this;
            this.$form.find('[data-formhall-constraints][data-formhall-constraints!=""]').each(function (index, formElement) {
                var $formElement = $(formElement);
                var constraints = $formElement.data('formhallConstraints');
                if (constraints) {
                    _this.adaptInputConstraints($formElement, constraints);
                }
            });
            this.$form.find('[data-formhall-container-constraints][data-formhall-container-constraints!=""]').each(function (index, container) {
                var $container = $(container);
                var constraints = $container.data('formhallContainerConstraints');
                if (constraints) { // set of radio buttons
                    $container.find('input[type="radio"]').each(function (index, radio) {
                        var $radio = $(radio);
                        _this.adaptInputConstraints($radio, constraints);
                    });
                }
            });
        };
        FormHallManager.prototype.adaptInputConstraints = function ($input, constraints) {
            if ($input.attr('type') === 'range') {
                constraints.min = constraints.min || 0;
                constraints.max = constraints.max || 100;
            }
            if (constraints.min) {
                $input.attr('min', constraints.min);
            }
            if (constraints.max) {
                $input.attr('max', constraints.max);
            }
            if (constraints.required) {
                $input.attr('required', constraints.required);
            }
            if (constraints.pattern) {
                $input.attr('pattern', constraints.pattern);
            }
            if (constraints.notBlank) {
                $input.attr('data-rule-notBlank', 'true'); // Do not use $.data() as it would break notBlank to not-blank when translating from camel case
                $input.attr('data-msg-notBlank', window['messages'] ? window['messages']['messages']['FH_FIELD_NOT_BLANK'] : 'This field must not be blank.');
            }
            if (constraints.minLength) {
                $input.attr('minlength', constraints.minLength);
            }
            if (constraints.maxLength) {
                $input.attr('maxlength', constraints.maxLength);
            }
            // See https://developer.mozilla.org/fr/docs/Web/Guide/HTML/HTML5/Constraint_validation for HTML5 capabilities
            // See http://johnnycode.com/2014/03/27/using-jquery-validate-plugin-html5-data-attribute-rules/ for available rules
        };
        FormHallManager.prototype.adaptInputRanges = function () {
            var _this = this;
            this.$form.find('[data-formhall-role="field-range"]').each(function (index, container) {
                var $container = $(container);
                var $output = $container.find('output');
                var $input = $container.find('input');
                var $min = $container.find('[data-formhall-role="field-range-min-bound"]');
                var $max = $container.find('[data-formhall-role="field-range-max-bound"]');
                var defaultValue = $container.data('formhallContainerDefaultValue');
                var min = $input.attr('min') && parseInt($input.attr('min')) || 0;
                var max = $input.attr('max') && parseInt($input.attr('max')) || 100;
                $input.attr('value', typeof defaultValue === 'number' ? defaultValue : (min + max) / 2);
                $min.text(min);
                $max.text(max);
                var onChangeValue = function () {
                    $output.html($input.val().toString());
                };
                onChangeValue();
                $input.on('input', onChangeValue).on('change', onChangeValue);
                _this.addUpdateDataListener(onChangeValue);
            });
        };
        FormHallManager.prototype.setupChoiceItemsUrls = function () {
            var _this = this;
            this.$form.find('[data-formhall-items-url][data-formhall-items-url!=""]').each(function (index, element) {
                var $element = $(element);
                var itemsUrl = $element.data('formhallItemsUrl');
                if (itemsUrl) {
                    _this.onShowListeners.push(function () {
                        _this.populateChoiceItemsFrom($element, itemsUrl, $element.prop('tagName') !== 'DATALIST');
                    });
                }
            });
        };
        FormHallManager.prototype.populateChoiceItemsFrom = function ($element, url, keepFirst) {
            var _this = this;
            // This is ok for select and suggestions for text-like fields, but not allowed for radio
            $.ajax({
                context: this,
                url: url,
                dataType: 'json',
                headers: {
                    'Cache-Control': 'max-age=0'
                },
                success: function (data) {
                    var $options = $element.find('option');
                    if (keepFirst) {
                        // Keep the first one that contains the label 
                        $options.slice(1).remove();
                    }
                    else {
                        $options.remove();
                    }
                    for (var _i = 0, data_1 = data; _i < data_1.length; _i++) {
                        var item = data_1[_i];
                        $element.append('<option value="' + (item.value ? item.value : '') + '">' + item.label + '</option>');
                    }
                    _this.onUpdateData();
                    var value = $element.attr('data-selected-value');
                    if (typeof value !== 'undefined') {
                        if ($element.prop('tagName') === 'SELECT') {
                            $element.find('option[value=' + value + ']').attr('selected', 'selected');
                        }
                    }
                    else if ($element.hasClass('readonly') || $element.is(":disabled")) {
                        // Select first by default
                        $element.find('option[value][value!=""]').first().attr('selected', 'selected');
                    }
                },
                error: function (request, textStatus, errorThrown) {
                    console.error(textStatus + ': ' + errorThrown);
                }
            });
        };
        FormHallManager.prototype.installDepends = function () {
            var _this = this;
            var baseId = this.$form.data('formhallDependsBaseId');
            this.$form.find('[data-formhall-depends][data-formhall-depends!=""]').each(function (index, container) {
                var $container = $(container);
                var depends = $container.data('formhallDepends');
                var dependsSel = '[id="' + baseId + depends.name + '"]';
                var dependsSelStart = '[id^="' + baseId + depends.name + '"]';
                var toggleFunction = function () {
                    _this.toggleDependency(dependsSel, $container, depends);
                };
                $(dependsSelStart).change(toggleFunction);
                _this.addUpdateDataListener(toggleFunction);
            });
        };
        FormHallManager.prototype.installValidate = function () {
            var _this = this;
            // See also: https://jqueryvalidation.org/documentation/#link-list-of-built-in-validation-methods
            var errorClass = 'text-danger';
            var $form = this.$form;
            this.validator = $form.validate({
                submitHandler: function (form) {
                    var readOnlyFields = $(form).find('input[name][readonly=readonly],textarea[name][readonly=readonly],select[name][readonly=readonly]');
                    readOnlyFields.removeAttr('disabled'); // To make sure those fields are part of the submit
                    _this.submitHandler(form);
                    readOnlyFields.attr('disabled', 'disabled');
                },
                errorPlacement: function (error, element) {
                    if (element.is('input:checkbox')) {
                        error.addClass('error-checkbox');
                        var parent = element.parent();
                        parent.append(error);
                    }
                    else if (element.is('input:radio')) {
                        error.addClass('error-radio-select');
                        var last = element.parent().children(':last');
                        error.insertAfter(last);
                    }
                    else if (element.is('select')) {
                        error.addClass('error-radio-select');
                        error.insertAfter(element);
                    }
                    else {
                        error.insertAfter(element);
                    }
                },
                errorClass: errorClass
            });
            this.validator['hideAllErrors'] = function (form) {
                $form.find('label.' + errorClass).hide();
                $form.find('.' + errorClass).removeClass(errorClass);
            };
        };
        FormHallManager.prototype.onShow = function () {
            this.setDirty(this.isCreateForm());
            for (var _i = 0, _a = this.onShowListeners; _i < _a.length; _i++) {
                var onShowListener = _a[_i];
                onShowListener();
            }
        };
        FormHallManager.prototype.onShown = function () {
        };
        FormHallManager.prototype.onHide = function () {
        };
        FormHallManager.prototype.onHidden = function () {
            this.hideMessageAreas();
            this.$form[0].reset();
        };
        FormHallManager.prototype.isCreateForm = function () {
            return this.formSelector.indexOf('CreateForm') >= 0;
        };
        FormHallManager.prototype.onSubmitStart = function () {
            this.$form.find('[data-formhall-role="progress"]').toggleClass('hidden', false);
        };
        FormHallManager.prototype.onSubmitComplete = function () {
            this.$form.find('[data-formhall-role="progress"]').toggleClass('hidden', true);
        };
        FormHallManager.prototype.submitHandler = function (form) {
            this.disableSubmitReset(true);
            this.setDirty(false);
            this.hideMessageAreas();
            this.onSubmitStart();
            var data = $(form).serializeJSON({
            //useIntKeysAsArrayIndex: true, // TODO: use this when we'll support collections
            });
            // Allow customization of submit request
            if (this.handlers && typeof this.handlers.customSubmit === 'function') {
                this.handlers.customSubmit.call(this, data);
            }
            else {
                this.defaultSubmit(data);
            }
        };
        // Default submit produces an Ajax json request which response is a simple string (with success/failure in status) to be displayed to the end-user
        FormHallManager.prototype.defaultSubmit = function (data) {
            var _this = this;
            var json = JSON.stringify(data);
            $.ajax({
                context: this,
                url: this.$form.attr('action'),
                method: this.$form.attr('method'),
                contentType: "application/json; charset=utf-8",
                data: json,
                statusCode: {
                    200: function (data, textStatus, request) {
                        // Successful edit
                        this.onSuccess(request.responseText);
                        this.updateFrom(request.getResponseHeader('Location'));
                    },
                    201: function (data, textStatus, request) {
                        // Successful add
                        this.onSuccess(request.responseText, request);
                        this.updateFrom(request.getResponseHeader('Location'));
                    },
                    400: function (request, textStatus, errorThrown) {
                        // Wrong data
                        this.onError(request.responseText);
                    },
                    405: function (request, textStatus, errorThrown) {
                        // Method not supported: we're in test environment
                        // Let's share the submitted data with HtmlUnit
                        window['formHallDebugSubmittedData'] = json;
                        //console.log(window.formHallDebugSubmittedData);
                    },
                    409: function (request, textStatus, errorThrown) {
                        var _this = this;
                        // Conflict
                        this.onError(request.responseText);
                        // TODO: would be nicer to have a link or button in the error area to refresh data from server
                        var msg = window['messages'] ?
                            window['messages']['messages']['FH_CONFIRM_REFRESH'] :
                            'Do you want to refresh the form with the latest version?\nNote: This will discard your local changes.';
                        if (confirm(msg)) {
                            this.updateFrom(request.getResponseHeader('Location'), function () {
                                _this.hideMessageAreas();
                            });
                        }
                    }
                },
                complete: function () {
                    _this.onSubmitComplete();
                }
            });
        };
        FormHallManager.prototype.hideMessageAreas = function () {
            this.$form.find('[data-formhall-role="success-area"]').toggleClass('hidden', true);
            this.$form.find('[data-formhall-role="error-area"]').toggleClass('hidden', true);
        };
        FormHallManager.prototype.onSuccess = function (message, request) {
            this.$form.find('[data-formhall-role="success-area-message"]').html(message);
            this.$form.find('[data-formhall-role="success-area"]').removeClass('hidden');
            if (this.handlers && typeof this.handlers.whenSuccess === 'function') {
                this.handlers.whenSuccess(request);
            }
        };
        FormHallManager.prototype.onError = function (message) {
            this.$form.find('[data-formhall-role="error-area-message"]').html(message);
            this.$form.find('[data-formhall-role="error-area"]').removeClass('hidden');
            if (this.handlers && typeof this.handlers.whenError === 'function') {
                this.handlers.whenError();
            }
        };
        FormHallManager.prototype.addUpdateDataListener = function (handler) {
            this.updateDataListeners.push(handler);
        };
        FormHallManager.prototype.onUpdateData = function () {
            for (var _i = 0, _a = this.updateDataListeners; _i < _a.length; _i++) {
                var updateDataListener = _a[_i];
                updateDataListener();
            }
        };
        FormHallManager.prototype.updateFrom = function (location, onSuccess, onError) {
            var _this = this;
            if (location) {
                $.ajax({
                    context: this,
                    url: location,
                    dataType: 'json',
                    headers: {
                        'Cache-Control': 'max-age=0'
                    },
                    success: function (data) {
                        _this.keepMessages = true;
                        _this.populateWith(data);
                        if (onSuccess) {
                            onSuccess(data);
                        }
                    },
                    error: function (request, textStatus, errorThrown) {
                        if (onError) {
                            onError(request, textStatus, errorThrown);
                        }
                    }
                });
            }
        };
        FormHallManager.prototype.getValueForName = function (name, data) {
            var names = name.split(/[[\]]{1,2}/);
            if (names.length > 1 && names[names.length - 1] === '') {
                names = names.slice(0, -1);
            }
            var value = data;
            for (var i = 0; i < names.length && value; i++) {
                value = value[names[i]];
            }
            return value;
        };
        FormHallManager.prototype.populateWith = function (data) {
            var _this = this;
            var $form = this.$form;
            $form[0].reset();
            $form.find('input[name]').each(function (index, input) {
                var name = input.name;
                var value = _this.getValueForName(name, data);
                if (typeof value !== 'undefined') {
                    switch (input.type) {
                        case 'checkbox':
                            $(input).attr('checked', value);
                            break;
                        case 'radio':
                            if (value === input.value) {
                                $(input).attr('checked', 'checked');
                            }
                            break;
                        default:
                            $(input).attr('value', value);
                            break;
                    }
                }
            });
            $form.find('textarea[name]').each(function (index, textarea) {
                var name = textarea.name;
                var value = _this.getValueForName(name, data);
                if (typeof value !== 'undefined') {
                    $(textarea).val(value);
                }
            });
            $form.find('select[name]').each(function (index, select) {
                var name = select.name;
                var value = _this.getValueForName(name, data);
                if (typeof value !== 'undefined') {
                    $(select).attr('data-selected-value', value);
                    $(select).find('option[value=' + value + ']').attr('selected', 'selected');
                }
            });
        };
        FormHallManager.prototype.toggleDependency = function (dependsSel, $container, depends) {
            var toggle = this.interpretSelector(dependsSel, depends.selector);
            if (depends.behavior === 'SHOW_HIDE') {
                $container.toggleClass('hidden', !toggle);
            }
            else if (depends.behavior === 'ENABLE_DISABLE') {
                $container.toggleClass('disabled', !toggle);
            }
            else if (depends.behavior === 'READWRITE_READONLY') {
                $container.toggleClass('readonly', !toggle);
            }
            var includedFields = $container.find('input[name],textarea[name],select[name]');
            // TODO: this may fail in case of any dependent inner group in this dependent inner group
            if (depends.behavior === 'READWRITE_READONLY') {
                if (toggle) {
                    includedFields.removeAttr('readonly');
                }
                else {
                    includedFields.attr('readonly', 'readonly');
                }
            }
            if (toggle) {
                includedFields.removeAttr('disabled');
            }
            else {
                includedFields.attr('disabled', 'disabled');
            }
        };
        FormHallManager.prototype.interpretSelector = function (dependsSel, selector) {
            var $property = $(dependsSel);
            var lparIndex = selector.indexOf('(');
            if (lparIndex >= 0) {
                return this.interpretFunctionalSelector($property, selector, lparIndex);
            }
            var selectors = typeof selector === 'undefined' ? [] : selector.split('|');
            if (!$property.length) {
                // Radio with enums
                // as given dependsSel doesn't exist as such, try with _<selector> suffices as they are in radio buttons
                var isChecked = false;
                var standardSelector = dependsSel.charAt(0) === '#';
                var insertIndex = standardSelector ? 0 : dependsSel.length - 2;
                for (var i = 0; i < selectors.length; i++) {
                    $property = standardSelector ?
                        $(dependsSel + '_' + selectors[i]) :
                        $(dependsSel.substring(0, insertIndex) + '_' + selectors[i] + dependsSel.substring(insertIndex));
                    if ($property.length && $property.is(':checked')) {
                        isChecked = true;
                        break;
                    }
                }
                return isChecked;
            }
            else if ($property.prop('tagName') === 'INPUT' && $property.prop('type') === 'checkbox') {
                // Checkbox
                var cssSelector = null;
                if (selector === 'false') {
                    cssSelector = ':unchecked';
                }
                else {
                    cssSelector = ':checked';
                }
                return $property.is(cssSelector);
            }
            else if ($property.prop('tagName') === 'SELECT') {
                // Select with enums
                var isSelected = false;
                var standardSelector = dependsSel.charAt(0) === '#';
                var insertIndex = standardSelector ? 0 : dependsSel.length - 2;
                for (var _i = 0, selectors_1 = selectors; _i < selectors_1.length; _i++) {
                    var selector_1 = selectors_1[_i];
                    $property = standardSelector ?
                        $(dependsSel + '_' + selector_1) :
                        $(dependsSel.substring(0, insertIndex) + '_' + selector_1 + dependsSel.substring(insertIndex));
                    if ($property.length && $property.prop('selected')) {
                        isSelected = true;
                        break;
                    }
                }
                return isSelected;
                //	  } else {
                // TODO: all other fields
            }
            return true;
        };
        FormHallManager.prototype.interpretFunctionalSelector = function ($property, selector, lparIndex) {
            var functionName = selector.substring(0, lparIndex);
            var args = selector.substring(lparIndex + 1, selector.length - 1).split(',');
            if (functionName === 'minOptions') {
                var minValue = new Number(args[0]);
                if ($property.prop('tagName') === 'SELECT') {
                    // Select
                    return $property.find('option[value][value!=""]').length >= minValue;
                }
                // Add other functional dependency selectors here...
            }
        };
        FormHallManager.prototype.disableSubmitReset = function (disabled) {
            $(this.formSelector + ' [type=submit], ' + this.formSelector + ' [type=reset]').prop('disabled', disabled);
        };
        ;
        FormHallManager.prototype.setDirty = function (dirty) {
            this.disableSubmitReset(!dirty);
            this.dirty = dirty;
        };
        ;
        return FormHallManager;
    }());
    exports.FormHallManager = FormHallManager;
    $.validator.addMethod("notBlank", function (value, element, param) {
        if (this.optional(element)) {
            return true;
        }
        return value.trim().length > 0;
    }, "This field must not be blank."); // Replaced with correct L10N set in data-msg-notBlank wherever needed
});
