/****************************************************************
 Licensed Materials - Property of IBM and/or HCL
(c) Copyright HCL Technologies Ltd. 2018.  All Rights Reserved.

 Note to U.S. Government Users Restricted Rights:
 Use, duplication or disclosure restricted by GSA ADP Schedule:
 Contract with IBM Corp.
*****************************************************************/
define(["require", "exports", "jquery", "d3"], function (require, exports, $, d3) {
    "use strict";
    exports.__esModule = true;
    function copyAndPush(array, newItem) {
        if (Array.isArray(array)) {
            var i = array.length;
            var newArray = Array(i + 1);
            while (i--) {
                newArray[i] = array[i];
            }
            newArray[array.length] = newItem;
            return newArray;
        }
    }
    function visitObservableTree(observable, onLeaf, parents) {
        if (observable) {
            if (observable['@type'] === 'RMObservableContainer' && observable.children && observable.children.length) {
                var newParents = copyAndPush(parents, observable);
                observable.children.forEach(function (child, observable) {
                    visitObservableTree(child, onLeaf, newParents);
                });
            }
            else {
                onLeaf(observable, parents);
            }
        }
    }
    function concatObservableParentDescriptions(parents, observableData) {
        // Always ignore parents[0]
        observableData.optionText = '';
        for (var i = 1; i < parents.length; i++) {
            observableData.optionText += parents[i].description + ' - ';
        }
        observableData.optionText += observableData.description;
    }
    function collectSelectedObservableDatas(sourceInfo) {
        var array = [];
        visitObservableTree(sourceInfo.observableRoot, function (observableData, parents) {
            if (sourceInfo.selectedObservableDatas.indexOf(observableData.id) >= 0) {
                concatObservableParentDescriptions(parents, observableData);
                array.push(observableData);
            }
        }, []);
        return array;
    }
    function mapSelectedObservableDatas(observableDatasArray) {
        var map = {};
        observableDatasArray.forEach(function (observableData) {
            map[observableData.id] = observableData;
        });
        return map;
    }
    var ObservationsChart = /** @class */ (function () {
        function ObservationsChart(sourceInfo, observationsUrl, $detailsContainer) {
            var _this = this;
            this.sourceInfo = sourceInfo;
            this.observationsUrl = observationsUrl;
            this.$detailsContainer = $detailsContainer;
            this.svg = null;
            this.stopped = true;
            this.restarted = false;
            $detailsContainer[0]['observations-chart'] = this;
            this.observableDatas = collectSelectedObservableDatas(sourceInfo);
            if (this.observableDatas.length == 0) {
                return;
            }
            var storedId = localStorage.getItem(this.sourceInfo.sourceId);
            //verify if saved id still exist in this source:
            if (storedId) {
                var found = false;
                for (var i = 0; i < this.observableDatas.length; i++) {
                    if (this.observableDatas[i].id === storedId) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    storedId = null;
                }
            }
            if (storedId) {
                this.selectedObservableId = storedId;
            }
            else {
                this.selectedObservableId = this.observableDatas[0].id; // There is always at least one (see configuredSource.html, data-template-partial="details-source")
            }
            this.observableDatasMap = mapSelectedObservableDatas(this.observableDatas);
            this.select = d3.select($detailsContainer[0]).select('select');
            this.select.attr('data-toggle', 'tooltip');
            this.select.selectAll("option")
                .remove()
                .data(this.observableDatas)
                .enter().append("option")
                .attr("value", function (d) { return d.id; })
                .attr('data-toggle', 'tooltip')
                .attr('data-container', function (d) { return '#observables-selector-tooltip-container-' + d.id; })
                .text(function (d) { return d.optionText; });
            $(this.select.node()).val(this.selectedObservableId).change();
            $(this.select.node()).on('change', function (event) {
                _this.onSelectUserEvent(event.target.value);
                return false;
            }).on('keyup', function (event) {
                _this.onSelectUserEvent(event.target.value);
                return false;
            });
            this.setTitle();
        }
        ObservationsChart.prototype.setTitle = function () {
            var _this = this;
            var expl = null;
            if (this.selectedObservableId && this.observableDatasMap[this.selectedObservableId]) {
                expl = this.observableDatasMap[this.selectedObservableId].explanation;
                if ("<lazy>" === expl) {
                    var id = this.observableDatasMap[this.selectedObservableId].id;
                    var lazyUrlStart = 'source-observables/' + this.sourceInfo.sourceId + '/';
                    //this from DialogManage .loadItem
                    $.ajax({
                        url: lazyUrlStart + id,
                        dataType: 'json',
                        headers: {
                            'Cache-Control': 'max-age=0'
                        },
                        success: function (data) {
                            // complete(data);
                            _this.observableDatasMap[_this.selectedObservableId].explanation = data.explanation;
                            //select may have changed during xhr ?
                            _this.updateTooltip(data.explanation);
                        },
                        error: function (xhr, textStatus, errorThrown) {
                            //TODO: better error reporting	                
                            console.log("Error: loading lazy ", textStatus, " ", errorThrown);
                        }
                    });
                    expl = null;
                }
            }
            this.updateTooltip(expl);
        };
        ObservationsChart.prototype.updateTooltip = function (expl) {
            var selected = this.select.selectAll('option').filter(function (d, i) { return this.selected; });
            var text = selected.text() + (expl ? (": " + expl) : "");
            //tooltip must be set on select, set on option it never appear
            this.select.attr('title', text);
            $(this.select.node()).tooltip('hide');
            $(this.select.node()).tooltip('fixTitle');
        };
        ObservationsChart.prototype.onSelectUserEvent = function (selectedObservableId) {
            if (selectedObservableId !== this.selectedObservableId) {
                this.selectedObservableId = selectedObservableId;
                localStorage.setItem(this.sourceInfo.sourceId, selectedObservableId);
                this.setTitle();
                this.displayData();
            }
        };
        ObservationsChart.prototype.startDisplay = function () {
            var _this = this;
            if (this.noLongerUsable || !this.checkUsable()) {
                return false;
            }
            if (this.stopped) {
                if (!this.svg) {
                    this.svg = d3.select(this.$detailsContainer[0]).select('svg');
                }
                this.svg.selectAll('*').remove();
                var bbox = this.svg.node().getBoundingClientRect();
                this.width = bbox.width;
                this.height = bbox.height;
                this.margin = { top: 20, right: 0, bottom: 30, left: 50 };
                this.g = this.svg.append('g'); // translate will be done later when knowing width of axisLeft
                this.x = d3.scaleTime(); // rangeRound will be done later when knowing width of axisLeft
                this.y = d3.scaleLinear()
                    .rangeRound([this.height - this.margin.top - this.margin.bottom, 0]);
                this.area = d3.area()
                    .x(function (d) { return _this.x(d.date); })
                    .y0(this.height - this.margin.top - this.margin.bottom)
                    .y1(function (d) { return _this.y(d.value); })
                    .defined(function (d) { return !d.last; });
                this.line = d3.line()
                    .x(function (d) { return _this.x(d.date); })
                    .y(function (d) { return _this.y(d.value); })
                    .defined(function (d) { return !d.last; });
                this.stopped = false;
                this.restarted = true;
                if (!this.observationsMap) {
                    this.gatherObservations(1000);
                }
                else {
                    this.displayData();
                    this.armGatherTimer(1000, 1000);
                }
            }
            return true;
        };
        ObservationsChart.prototype.gatherObservations = function (delay) {
            var _this = this;
            if (!this.checkUsable()) {
                this.stopDisplay();
                return;
            }
            var startTime = new Date().getTime();
            // 50 millis margin is to provision for setTimeout() inaccuracy
            $.getJSON(this.observationsUrl + (typeof delay === 'number' && delay > 50 ? '?maxAgeMillis=' + (delay - 50) : ''))
                .done(function (observations, textStatus, jqXHR) {
                var responseTime = new Date().getTime();
                var serviceClockString = jqXHR.getResponseHeader('X-RMS-Service-Clock');
                if (serviceClockString) {
                    var serviceClock = Number(serviceClockString);
                    window['serviceClockSkew'] = new Date().getTime() - serviceClock;
                }
                var serviceClockSkew = (typeof window['serviceClockSkew'] === 'number' ? window['serviceClockSkew'] : 0);
                _this.sourceInfo.lastObservedOn = observations.timestamp + serviceClockSkew;
                _this.populateWith(observations, new Date(observations.timestamp + serviceClockSkew));
                if (!_this.stopped) {
                    var nextTick = startTime + delay - responseTime;
                    _this.armGatherTimer(delay, nextTick);
                }
            })
                .fail(function (jqxhr, textStatus, error) {
                _this.stopDisplay();
            });
        };
        ObservationsChart.prototype.checkUsable = function () {
            if (this.$detailsContainer.length === 0 || this.$detailsContainer.closest('body').length === 0 || this.$detailsContainer[0]['observations-chart'] !== this) {
                // In case we've been removed, after an Ajax refresh, then stop gathering data
                this.noLongerUsable = true;
                return false;
            }
            return true;
        };
        ObservationsChart.prototype.armGatherTimer = function (delay, nextTick) {
            var _this = this;
            this.gatherTimer = setTimeout(function () {
                _this.gatherObservations(delay);
            }, nextTick < 0 ? 0 : nextTick);
        };
        ObservationsChart.prototype.populateWith = function (observations, date) {
            var _this = this;
            if (!this.observationsMap) {
                this.observationsMap = {};
                if (!this.observableDatasMap) {
                    this.observableDatasMap = mapSelectedObservableDatas(this.observableDatas);
                }
            }
            $.each(observations.datas, function (id, value) {
                if (!_this.observationsMap[id]) {
                    _this.observationsMap[id] = [];
                    if (!_this.observableDatasMap || !_this.observableDatasMap[id]) {
                        _this.observableDatasMap = mapSelectedObservableDatas(_this.observableDatas);
                    }
                    if (!_this.observableDatasMap[id]) {
                        return;
                    }
                    if (_this.observableDatasMap[id].specialHandling === 'DELTA') {
                        return; // Ignore first data
                    }
                }
                var ary = _this.observationsMap[id];
                var last = date.getTime();
                // insert a line/area break ?
                if (_this.restarted && ary.length > 0 && !ary[ary.length - 1].last) { //only if restarted and break is not yet defined
                    if ((last - ary[ary.length - 1].date.getTime()) > 2000) {
                        //add special entry (no date, no value) to start a new curve(area) segment
                        ary.push({ last: true }); ///insert a break in line/area
                        if (_this.observableDatasMap[id].specialHandling === 'DELTA') {
                            return; // Ignore first data (of new range)
                        }
                    }
                }
                ary.push({ date: date, value: value });
                //find data to be deleted due to max time range (0,1 or more value to delete)
                var index = -1;
                var i = 0;
                while (i < ary.length) {
                    if (!ary[i].last) { // if .last -> remove it
                        var first = ary[i].date.getTime();
                        var delta = last - first;
                        if (delta > 180000) { ///3mn in ms
                            index = i;
                        }
                        else {
                            break;
                        }
                    }
                    i++;
                }
                if (index >= 0) { //have data to remove ? 
                    var num = index - 0 + 1;
                    ary.splice(0, num); //remove first
                }
            });
            this.restarted = false;
            this.displayData();
        };
        ObservationsChart.prototype.updateWidth = function () {
            if (this.observationsMap) {
                this.displayData(); // to update width of chart
            }
        };
        ObservationsChart.prototype.displayData = function () {
            //width may have changed due to layout changed or browser width changed
            //note: height of chart is constant (see css)
            var bbox = this.svg.node().getBoundingClientRect();
            this.width = bbox.width;
            var data = this.observationsMap[this.selectedObservableId];
            if (!data || !data.length) {
                this.selectedObservableId = this.observableDatas[0].id;
                data = this.observationsMap[this.selectedObservableId];
                if (!data || !data.length) {
                    return;
                }
            }
            //take care of special value (.last==true) => no date or value available
            var minTime, maxTime;
            var i_maxTime = -1;
            if (data.length == 0) {
                minTime = 0;
                maxTime = 1000; //ms
            }
            else {
                var i_minTime = -1;
                for (var i = 0; i < data.length; i++) {
                    if (!data[i].last) {
                        i_minTime = i;
                        break;
                    }
                }
                for (var i = data.length - 1; i >= 0; i--) {
                    if (!data[i].last) {
                        i_maxTime = i;
                        break;
                    }
                }
                if (i_minTime < 0) { //no data
                    minTime = 0;
                    maxTime = 1000; //ms
                }
                else {
                    minTime = data[i_minTime].date.getTime();
                    maxTime = data[i_maxTime].date.getTime();
                }
            }
            //var minTime = data[0].date.getTime() - (data.length > 1 ? 0 : 1000);
            //var maxTime = data[data.length-1].date.getTime();
            var xDelta = (maxTime - minTime) * 0.01;
            this.x.domain([new Date(minTime - xDelta), new Date(maxTime + xDelta)]);
            var specialHandling = this.observableDatasMap[this.selectedObservableId].specialHandling;
            var yExtent = d3.extent(data, function (d) {
                return typeof d.value === 'number' ? d.value : undefined;
            });
            var minValue = yExtent[0];
            var maxValue = yExtent[1];
            if (specialHandling !== 'PERCENT') {
                if (maxValue === minValue) {
                    if (minValue !== 0) {
                        this.y.domain([minValue > 0 ? 0 : minValue, minValue < 0 ? 0 : minValue]);
                    }
                    else {
                        this.y.domain([0, 1]);
                    }
                }
                else {
                    var yDelta = (maxValue - minValue) * 0.05;
                    this.y.domain([minValue - yDelta, maxValue + yDelta]);
                }
            }
            else {
                this.y.domain([0, Math.max(100, maxValue)]);
            }
            this.g.selectAll('*').remove();
            var g = this.g.append('g');
            var leftAxis = this.g.append('g')
                .call(d3.axisLeft(this.y));
            var leftAxisMaxWidth = 0;
            leftAxis.selectAll('text')
                .each(function (d, i) {
                var self = d3.select(this);
                var thisWidth = this.getComputedTextLength() - Number(self.attr('x'));
                if (leftAxisMaxWidth < thisWidth) {
                    leftAxisMaxWidth = thisWidth;
                }
            });
            this.margin.left = leftAxisMaxWidth;
            this.g.attr('transform', 'translate(' + this.margin.left + ',' + this.margin.top + ')');
            var x_width = this.width - this.margin.left - this.margin.right - 1;
            this.x.rangeRound([0, x_width]);
            g.attr('transform', 'translate(0,' + (this.height - this.margin.top - this.margin.bottom) + ')')
                .call(d3.axisBottom(this.x));
            this.g.append('path')
                .datum(data)
                .attr('fill', '#CDF')
                .attr('stroke-width', 0)
                .attr('d', this.area);
            this.g.append('path')
                .datum(data)
                .attr('fill', 'none')
                .attr('stroke', '#46D')
                .attr('stroke-width', 2)
                .attr('d', this.line);
            //text on top/right of graph
            if (i_maxTime >= 0) { //exists
                var value = data[i_maxTime].value;
                var text = value;
                if (typeof value === 'number') {
                    var unit = this.observableDatasMap[this.selectedObservableId].unit;
                    var val = this.roundValue(value);
                    var format = d3.format(",");
                    text = format(val);
                    if (unit)
                        text = text + ' ' + unit;
                }
                leftAxis.append('text')
                    .attr('fill', typeof value === 'number' ? '#46D' : '#D64')
                    .attr('x', x_width)
                    .attr('y', 0)
                    .attr('dy', '-0.6em')
                    .attr('text-anchor', 'end')
                    .text(text);
            }
        };
        //TODO: why not use d3.round() ??
        ObservationsChart.prototype.roundValue = function (value) {
            var log10 = Math.round(Math.log(value) * Math.LOG10E);
            var fixes = log10 < 0 ? 4 - log10 : Math.max(0, 5 - log10);
            fixes = Math.max(20, fixes);
            fixes = Math.min(0, fixes);
            return Number((value).toFixed(fixes));
        };
        ObservationsChart.prototype.stopDisplay = function () {
            if (!this.stopped) {
                if (this.gatherTimer) {
                    clearTimeout(this.gatherTimer);
                    this.gatherTimer = null;
                }
                this.stopped = true;
            }
        };
        return ObservationsChart;
    }());
    exports.ObservationsChart = ObservationsChart;
});
